@@ -89,8 +89,7 @@ int H3_EXPORT(getBaseCellNumber)(H3Index h) { return H3_GET_BASE_CELL(h); }
 /**
  * Converts a string representation of an H3 index into an H3 index.
  * @param str The string representation of an H3 index.
- * @return The H3 index corresponding to the string argument, or H3_NULL if
- * invalid.
+ * @param out Output: The H3 index corresponding to the string argument
  */
 H3Error H3_EXPORT(stringToH3)(const char *str, H3Index *out) {
     H3Index h = H3_NULL;
@@ -120,50 +119,184 @@ H3Error H3_EXPORT(h3ToString)(H3Index h, char *str, size_t sz) {
     return E_SUCCESS;
 }
 
+/*
+The top 8 bits of any cell should be a specific constant:
+
+- The 1 high bit should be `0`
+- The 4 mode bits should be `0001` (H3_CELL_MODE)
+- The 3 reserved bits should be `000`
+
+In total, the top 8 bits should be `0_0001_000`
+*/
+static inline bool _hasGoodTopBits(H3Index h) {
+    h >>= (64 - 8);
+    return h == 0b00001000;
+}
+
+/* Check that no digit from 1 to `res` is 7 (INVALID_DIGIT).
+
+MHI = 0b100100100100100100100100100100100100100100100;
+MLO = MHI >> 2;
+
+|  d  | d & MHI |  ~d | ~d - MLO | d & MHI & (~d - MLO) |  result |
+|-----|---------|-----|----------|----------------------|---------|
+| 000 |     000 |     |          |                  000 | OK      |
+| 001 |     000 |     |          |                  000 | OK      |
+| 010 |     000 |     |          |                  000 | OK      |
+| 011 |     000 |     |          |                  000 | OK      |
+| 100 |     100 | 011 | 010      |                  000 | OK      |
+| 101 |     100 | 010 | 001      |                  000 | OK      |
+| 110 |     100 | 001 | 000      |                  000 | OK      |
+| 111 |     100 | 000 | 111*     |                  100 | invalid |
+
+  *: carry happened
+
+
+Note: only care about identifying the *lowest* 7.
+
+Examples with multiple digits:
+
+|    d    | d & MHI |    ~d   | ~d - MLO | d & MHI & (~d - MLO) |  result |
+|---------|---------|---------|----------|----------------------|---------|
+| 111.111 | 100.100 | 000.000 | 110.111* |              100.100 | invalid |
+| 110.111 | 100.100 | 001.000 | 111.111* |              100.100 | invalid |
+| 110.110 | 100.100 | 001.001 | 000.000  |              000.000 | OK      |
+
+  *: carry happened
+
+In the second example with 110.111, we "misidentify" the 110 as a 7, due
+to a carry from the lower bits. But this is OK because we correctly
+identify the lowest (only, in this example) 7 just before it.
+
+We only have to worry about carries affecting higher bits in the case of
+a 7; all other digits (0--6) don't cause a carry when computing ~d - MLO.
+So even though a 7 can affect the results of higher bits, this is OK
+because we will always correctly identify the lowest 7.
+
+For further notes, see the discussion here:
+https://github.com/uber/h3/pull/496#discussion_r795851046
+*/
+static inline bool _hasAny7UptoRes(H3Index h, int res) {
+    const uint64_t MHI = 0b100100100100100100100100100100100100100100100;
+    const uint64_t MLO = MHI >> 2;
+
+    int shift = 3 * (15 - res);
+    h >>= shift;
+    h <<= shift;
+    h = (h & MHI & (~h - MLO));
+
+    return h != 0;
+}
+
+/* Check that all unused digits after `res` are set to 7 (INVALID_DIGIT).
+
+Bit shift to avoid looping through digits.
+*/
+static inline bool _hasAll7AfterRes(H3Index h, int res) {
+    // NOTE: res check is needed because we can't shift by 64
+    if (res < 15) {
+        int shift = 19 + 3 * res;
+
+        h = ~h;
+        h <<= shift;
+        h >>= shift;
+
+        return h == 0;
+    }
+    return true;
+}
+
+/*
+Get index of first nonzero bit of an H3Index.
+
+When available, use compiler intrinsics, which should be fast.
+If not available, fall back to a loop.
+*/
+static inline int _firstOneIndex(H3Index h) {
+#if defined(__GNUC__) || defined(__clang__)
+    return 63 - __builtin_clzll(h);
+#elif defined(_MSC_VER) && defined(_M_X64)  // doesn't work on win32
+    unsigned long index;
+    _BitScanReverse64(&index, h);
+    return (int)index;
+#else
+    // Portable fallback
+    int pos = 63 - 19;
+    H3Index m = 1;
+    while ((h & (m << pos)) == 0) pos--;
+    return pos;
+#endif
+}
+
+/*
+One final validation just for cells whose base cell (res 0)
+is a pentagon.
+
+Pentagon cells start with a sequence of 0's (CENTER_DIGIT's).
+The first nonzero digit can't be a 1 (i.e., "deleted subsequence",
+PENTAGON_SKIPPED_DIGIT, or K_AXES_DIGIT).
+
+We can check that (in the lower 45 = 15*3 bits) the position of the
+first 1 bit isn't divisible by 3.
+*/
+static inline bool _hasDeletedSubsequence(H3Index h, int base_cell) {
+    // TODO: https://github.com/uber/h3/issues/984
+    static const bool isBaseCellPentagonArr[128] = {
+        [4] = 1,  [14] = 1, [24] = 1, [38] = 1, [49] = 1,  [58] = 1,
+        [63] = 1, [72] = 1, [83] = 1, [97] = 1, [107] = 1, [117] = 1};
+
+    if (isBaseCellPentagonArr[base_cell]) {
+        h <<= 19;
+        h >>= 19;
+
+        if (h == 0) return false;  // all zeros: res 15 pentagon
+        return _firstOneIndex(h) % 3 == 0;
+    }
+    return false;
+}
+
 /**
  * Returns whether or not an H3 index is a valid cell (hexagon or pentagon).
  * @param h The H3 index to validate.
  * @return 1 if the H3 index if valid, and 0 if it is not.
  */
 int H3_EXPORT(isValidCell)(H3Index h) {
-    if (H3_GET_HIGH_BIT(h) != 0) return 0;
+    /*
+    Look for bit patterns that would disqualify an H3Index from
+    being valid. If identified, exit early.
 
-    if (H3_GET_MODE(h) != H3_CELL_MODE) return 0;
+    For reference the H3 index bit layout:
 
-    if (H3_GET_RESERVED_BITS(h) != 0) return 0;
+    |   Region   | # bits |
+    |------------|--------|
+    | High       |      1 |
+    | Mode       |      4 |
+    | Reserved   |      3 |
+    | Resolution |      4 |
+    | Base Cell  |      7 |
+    | Digit 1    |      3 |
+    | Digit 2    |      3 |
+    | ...        |    ... |
+    | Digit 15   |      3 |
 
-    int baseCell = H3_GET_BASE_CELL(h);
-    if (NEVER(baseCell < 0) || baseCell >= NUM_BASE_CELLS) {
-        // Base cells less than zero can not be represented in an index
-        return 0;
-    }
+    Speed benefits come from using bit manipulation instead of loops,
+    whenever possible.
+    */
+    if (!_hasGoodTopBits(h)) return false;
 
-    int res = H3_GET_RESOLUTION(h);
-    if (NEVER(res < 0 || res > MAX_H3_RES)) {
-        // Resolutions less than zero can not be represented in an index
-        return 0;
-    }
+    // No need to check resolution; any 4 bits give a valid resolution.
+    const int res = H3_GET_RESOLUTION(h);
 
-    bool foundFirstNonZeroDigit = false;
-    for (int r = 1; r <= res; r++) {
-        Direction digit = H3_GET_INDEX_DIGIT(h, r);
+    // Get base cell number and check that it is valid.
+    const int bc = H3_GET_BASE_CELL(h);
+    if (bc >= NUM_BASE_CELLS) return false;
 
-        if (!foundFirstNonZeroDigit && digit != CENTER_DIGIT) {
-            foundFirstNonZeroDigit = true;
-            if (_isBaseCellPentagon(baseCell) && digit == K_AXES_DIGIT) {
-                return 0;
-            }
-        }
+    if (_hasAny7UptoRes(h, res)) return false;
+    if (!_hasAll7AfterRes(h, res)) return false;
+    if (_hasDeletedSubsequence(h, bc)) return false;
 
-        if (NEVER(digit < CENTER_DIGIT) || digit >= NUM_DIGITS) return 0;
-    }
-
-    for (int r = res + 1; r <= MAX_H3_RES; r++) {
-        Direction digit = H3_GET_INDEX_DIGIT(h, r);
-        if (digit != INVALID_DIGIT) return 0;
-    }
-
-    return 1;
+    // If no disqualifications were identified, the index is a valid H3 cell.
+    return true;
 }
 
 /**
@@ -187,8 +320,7 @@ void setH3Index(H3Index *hp, int res, int baseCell, Direction initDigit) {
  *
  * @param h H3Index to find parent of
  * @param parentRes The resolution to switch to (parent, grandparent, etc)
- *
- * @return H3Index of the parent, or H3_NULL if you actually asked for a child
+ * @param out Output: H3Index of the parent
  */
 H3Error H3_EXPORT(cellToParent)(H3Index h, int parentRes, H3Index *out) {
     int childRes = H3_GET_RESOLUTION(h);
@@ -231,9 +363,8 @@ static bool _hasChildAtRes(H3Index h, int childRes) {
  *
  * @param h         H3Index to find the number of children of
  * @param childRes  The child resolution you're interested in
- *
- * @return int      Exact number of children (handles hexagons and pentagons
- *                  correctly)
+ * @param out      Output: exact number of children (handles hexagons and
+ * pentagons correctly)
  */
 H3Error H3_EXPORT(cellToChildrenSize)(H3Index h, int childRes, int64_t *out) {
     if (!_hasChildAtRes(h, childRes)) return E_RES_DOMAIN;
@@ -547,8 +678,8 @@ H3Error H3_EXPORT(compactCells)(const H3Index *h3Set, H3Index *compactedSet,
  *
  * Skips elements that are H3_NULL (i.e., 0).
  *
- * @param   compactSet  Set of compacted cells
- * @param   numCompact  The number of cells in the input compacted set
+ * @param   compactedSet  Set of compacted cells
+ * @param   numCompacted  The number of cells in the input compacted set
  * @param   outSet      Output array for decompressed cells (preallocated)
  * @param   numOut      The size of the output array to bound check against
  * @param   res         The H3 resolution to decompress to
@@ -577,7 +708,7 @@ H3Error H3_EXPORT(uncompactCells)(const H3Index *compactedSet,
  * the exact size of the uncompacted set of hexagons.
  *
  * @param   compactedSet  Set of hexagons
- * @param   numHexes      The number of hexes in the input set
+ * @param   numCompacted  The number of hexes in the input set
  * @param   res           The hexagon resolution to decompress to
  * @param   out           The number of hexagons to allocate memory for
  * @returns E_SUCCESS on success, or another value on error
@@ -1114,6 +1245,10 @@ static H3Error validateChildPos(int64_t childPos, H3Index parent,
 /**
  * Returns the position of the cell within an ordered list of all children of
  * the cell's parent at the specified resolution
+ * @param child Child cell index
+ * @param parentRes Resolution of the parent cell to find the position within
+ * @param out Output: The position of the child cell within its parents cell
+ * list of children
  */
 H3Error H3_EXPORT(cellToChildPos)(H3Index child, int parentRes, int64_t *out) {
     int childRes = H3_GET_RESOLUTION(child);
@@ -1189,7 +1324,12 @@ H3Error H3_EXPORT(cellToChildPos)(H3Index child, int parentRes, int64_t *out) {
 
 /**
  * Returns the child cell at a given position within an ordered list of all
- * children at the specified resolution */
+ * children at the specified resolution
+ * @param childPos Position within the ordered list
+ * @param parent Parent cell of the cell index to find
+ * @param childRes Resolution of the child cell index
+ * @param child Output: child cell index
+ */
 H3Error H3_EXPORT(childPosToCell)(int64_t childPos, H3Index parent,
                                   int childRes, H3Index *child) {
     // Validate resolution
